"""
$description Live TV channels and video on-demand service from TV Azteca, a Mexican free-to-air broadcaster.
$url tvazteca.com
$url adn40.mx
$type live, vod
$region Mexico
"""

import re
from dataclasses import asdict, dataclass
from urllib.parse import urlencode, urlparse

from streamlink import PluginError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import url_concat


log = getLogger(__name__)


@dataclass
class SafeModeTokenRequest:
    asset_id: str
    type: str
    token: str


@dataclass
class TVAztecaPage:
    video_id: str
    embed_url: str
    title: str | None = None
    channel: str | None = None
    program: str | None = None


@dataclass
class EnVivoConfig:
    client_id: str
    player_id: str
    app_version: str


@dataclass
class EnVivoPageMedia:
    id: str
    title: str
    purchased: int


@dataclass
class EnVivoMedia:
    id: str
    title: str
    access_token: str
    channel: str | None = None
    program: str | None = None


@dataclass
class EnVivoPlayback:
    hls_url: str
    uid: str
    sid: str
    pid: str


@pluginmatcher(
    name="tvazteca",
    pattern=re.compile(
        r"https?://(?:www\.)?tvazteca\.com/(?P<path>\S+?)/?$",
    ),
)
@pluginmatcher(
    name="adn40",
    pattern=re.compile(
        r"https?://(?:www|live)\.adn40\.mx(?:/(?P<path>\S*))?$",
    ),
)
@pluginmatcher(
    name="envivo_live",
    pattern=re.compile(
        r"https?://envivo\.tvazteca\.com/watch/live/(?P<id>[^/?#]+)/?$",
    ),
)
@pluginmatcher(
    name="envivo_vod",
    pattern=re.compile(
        r"https?://envivo\.tvazteca\.com/watch/media/(?P<id>[^/?#]+)/?$",
    ),
)
class TVAzteca(Plugin):
    _ENVIVO_DOMAIN = "envivo.tvazteca.com"
    _ENVIVO_ORIGIN = f"https://{_ENVIVO_DOMAIN}"
    _ENVIVO_REFERER = f"https://{_ENVIVO_DOMAIN}/"
    _ENVIVO_LIVE_QUERY = """
    query getPlayerLive($id: String!) {
      getLive(_id: $id) {
        _id
        name
        accessToken
        schedules {
          name
          current
        }
      }
    }
    """
    _ENVIVO_MEDIA_QUERY = """
    query getPlayerMedia($id: String!) {
      getMedia(_id: $id) {
        _id
        title
        accessToken
      }
    }
    """
    _ENVIVO_PAGE_MEDIA_QUERY = """
    query getPlayerPageMedia($id: String!) {
      getMedia(_id: $id) {
        _id
        title
        purchased
      }
    }
    """

    _ENVIVO_APP_NAME = "tv-azteca-en-vivo-app-web"
    _ENVIVO_DEFAULT_RESOLUTION = "1920x1080"

    _MDSTRM_ENDPOINT_ENVIVO_LIVE = "live-stream"
    _MDSTRM_ENDPOINT_ENVIVO_VOD = "video"

    _SCHEMA_PLAYER = validate.Schema(
        validate.xml_xpath(
            ".//div[@data-video-player][@data-video-id][@data-player-id][@data-video-type][1]",
        ),
        validate.none_or_all(
            validate.get(0),
            validate.union((
                validate.get("data-video-id"),
                validate.get("data-player-id"),
                validate.get("data-video-type"),
                validate.get("data-video-title"),
                validate.get("data-video-channel"),
                validate.get("data-video-program"),
                validate.all(
                    validate.get("data-safe-mode-config"),
                    validate.none_or_all(
                        validate.parse_json(),
                        {
                            "domain": str,
                            "bodyJson": validate.all(
                                str,
                                validate.parse_json(),
                                {
                                    "asset_id": str,
                                    "type": str,
                                    "token": str,
                                },
                                validate.union_get("asset_id", "type", "token"),
                            ),
                        },
                        validate.union_get("domain", "bodyJson"),
                    ),
                ),
            )),
        ),
    )
    _SCHEMA_SCRIPT = validate.Schema(
        validate.xml_xpath(".//script[@type='application/ld+json'][contains(text(),'VideoObject')]/text()"),
        validate.none_or_all(
            [
                validate.all(
                    validate.parse_json(),
                    validate.any(
                        list,
                        validate.transform(lambda obj: [obj]),
                    ),
                    [dict],
                    validate.filter(lambda obj: obj.get("@type", "") == "VideoObject"),
                    validate.get(0),
                    validate.none_or_all(
                        {
                            "embedUrl": validate.url(
                                hostname="mdstrm.com",
                            ),
                            "name": validate.any(None, str),
                            "mainEntityOfPage": {
                                "publisher": {
                                    "name": validate.any(None, str),
                                },
                            },
                        },
                        validate.union_get(
                            "embedUrl",
                            "name",
                            ("mainEntityOfPage", "publisher", "name"),
                        ),
                    ),
                ),
            ],
            validate.filter(bool),
            validate.get(0),
        ),
    )
    _SCHEMA_ADN_METADATA = validate.Schema(
        validate.union((
            validate.xml_xpath_string(
                ".//h1[contains(concat(' ', normalize-space(@class), ' '), ' b-adn-live-schedule__program-title ')][1]/text()",
            ),
            validate.xml_xpath_string(
                ".//meta[@name='datalayer-videotitle'][1]/@content",
            ),
        )),
    )
    _SCHEMA_FUSION_PLAYER = validate.Schema(
        validate.xml_xpath(
            ".//script[@id='fusion-metadata'][contains(text(),'Fusion.tree')][1]/text()",
        ),
        validate.none_or_all(
            validate.get(0),
            re.compile(r"Fusion\.tree\s*=\s*(?P<json>\{.+?});", re.DOTALL),
            validate.none_or_all(
                validate.get("json"),
                validate.parse_json(),
                {
                    "children": [
                        validate.all(
                            {
                                "children": [
                                    {
                                        "type": str,
                                        "props": dict,
                                    },
                                ],
                            },
                            validate.get("children"),
                            validate.filter(lambda obj: obj["type"] == "tva-mediastream-block/default"),
                            validate.get(0),
                            validate.none_or_all(
                                {
                                    "props": {
                                        "customFields": {
                                            "videoId": str,
                                            "playerId": str,
                                            "videoType": str,
                                        },
                                    },
                                },
                                validate.union_get(
                                    ("props", "customFields", "videoId"),
                                    ("props", "customFields", "playerId"),
                                    ("props", "customFields", "videoType"),
                                ),
                            ),
                        ),
                    ],
                },
                validate.get("children"),
                validate.filter(bool),
                validate.get(0),
            ),
        ),
    )

    def _get_streams(self):
        if self.matches["tvazteca"] or self.matches["adn40"]:
            return self._get_tvazteca_streams()
        elif self.matches["envivo_live"]:
            return self._get_envivo_streams(self._MDSTRM_ENDPOINT_ENVIVO_LIVE)
        elif self.matches["envivo_vod"]:
            return self._get_envivo_streams(self._MDSTRM_ENDPOINT_ENVIVO_VOD)

    # tvazteca/adn40
    def _get_tvazteca_streams(self):
        page = self._get_tvazteca_page()

        log.debug(f"Video ID: {page.video_id}")
        log.debug(f"Embed URL: {page.embed_url}")

        streams = self.session.streams(page.embed_url)

        self.id = page.video_id
        self.title = page.title
        self.author = page.channel
        self.category = page.program

        return streams

    def _get_tvazteca_page(self) -> TVAztecaPage:
        root = self.session.http.get(
            self.url,
            schema=validate.Schema(validate.parse_html()),
        )

        # www.tvazteca.com live/video
        if player := self._SCHEMA_PLAYER.validate(root):
            video_id, player_id, video_type, title, channel, program, safe_mode = player

            access_token = None
            if safe_mode:
                domain, (asset_id, token_type, token) = safe_mode
                access_token = self._request_access_token(
                    domain,
                    SafeModeTokenRequest(
                        asset_id=asset_id,
                        type=token_type,
                        token=token,
                    ),
                )

            return TVAztecaPage(
                video_id=video_id,
                embed_url=self._get_mdstrm_url(
                    video_id,
                    player_id,
                    video_type,
                    access_token,
                ),
                title=title,
                channel=channel,
                program=program,
            )

        # adn40 live
        if player := self._SCHEMA_FUSION_PLAYER.validate(root):
            video_id, player_id, video_type = player
            title, channel = self._SCHEMA_ADN_METADATA.validate(root)

            return TVAztecaPage(
                video_id=video_id,
                embed_url=self._get_mdstrm_url(
                    video_id,
                    player_id,
                    video_type,
                ),
                title=title,
                channel=channel,
            )

        # adn40 video
        if script := self._SCHEMA_SCRIPT.validate(root):
            embed_url, title, channel = script
            path = urlparse(embed_url).path.rstrip("/")

            if not path:
                raise PluginError(f"Invalid MDStrm embed URL: {embed_url}")

            return TVAztecaPage(
                video_id=path.rsplit("/", 1)[-1],
                embed_url=embed_url,
                title=title,
                channel=channel,
            )

        raise PluginError("Unable to locate video metadata")

    def _request_access_token(
        self,
        token_url: str,
        token_request: SafeModeTokenRequest,
    ) -> str:
        log.debug("Requesting access token")

        response = self.session.http.post(
            token_url,
            json=asdict(token_request),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "status": "OK",
                    "access_token": str,
                },
            ),
        )

        return response["access_token"]

    @staticmethod
    def _get_mdstrm_url(
        video_id: str,
        player_id: str,
        video_type: str,
        access_token: str | None = None,
    ) -> str:
        parsed_type = "live-stream" if video_type == "live" else "embed"

        params = {
            "player": player_id,
        }

        if access_token:
            params["access_token"] = access_token

        return f"https://mdstrm.com/{parsed_type}/{video_id}?{urlencode(params)}"

    # envivo
    def _get_envivo_streams(self, endpoint):
        stream_id = self.match["id"]

        config = self._get_envivo_config()
        log.debug(f"Player ID: {config.player_id}")

        if endpoint == self._MDSTRM_ENDPOINT_ENVIVO_LIVE:
            media = self._get_envivo_live_metadata(stream_id, config.client_id)
        elif endpoint == self._MDSTRM_ENDPOINT_ENVIVO_VOD:
            media = self._get_envivo_vod_metadata(stream_id, config.client_id)
        else:
            raise PluginError(f"Unsupported MDStrm endpoint: {endpoint}")

        if not media:
            return None

        playback = self._get_envivo_playback(
            endpoint=endpoint,
            media_id=media.id,
            player_id=config.player_id,
            app_version=config.app_version,
            access_token=media.access_token,
        )

        log.debug(f"Found HLS URL: {playback.hls_url}")

        params = {
            "dnt": "true",
            "player": config.player_id,
            "access_token": media.access_token,
            "uid": playback.uid,
            "sid": playback.sid,
            "pid": playback.pid,
            "an": self._ENVIVO_APP_NAME,
            "at": "web-app",
            "av": config.app_version,
            "sc": "0",
            "ref": self._ENVIVO_DOMAIN,
            "res": self._ENVIVO_DEFAULT_RESOLUTION,
            "ext_pb": "0",
            "CMCD": f'cid="{media.id}",mtp=500,ot=m,sf=h,sid="{playback.pid}",su',
        }

        streams = HLSStream.parse_variant_playlist(
            self.session,
            playback.hls_url,
            params=params,
            headers={
                "Referer": self._ENVIVO_REFERER,
            },
        )

        self.id = media.id
        self.title = media.title
        self.author = media.channel
        self.category = media.program

        return streams

    def _get_envivo_page_metadata(
        self,
        stream_id: str,
        client_id: str,
    ) -> EnVivoPageMedia | None:
        log.debug("Requesting page metadata")

        payload = [
            {
                "operationName": "getPlayerPageMedia",
                "variables": {
                    "id": stream_id,
                },
                "query": self._ENVIVO_PAGE_MEDIA_QUERY,
            },
        ]

        datatype, result = self._graphql(
            client_id,
            payload,
            {
                "data": {
                    "getMedia": validate.any(
                        None,
                        validate.all(
                            {
                                "_id": str,
                                "title": str,
                                "purchased": int,
                            },
                            validate.union_get(
                                "_id",
                                "title",
                                "purchased",
                            ),
                        ),
                    ),
                },
            },
            validate.get(("data", "getMedia")),
        )

        if datatype == "error":
            log.error(result)
            return None

        if result is None:
            log.error(f"No data found for media with ID {stream_id}")
            return None

        media_id, title, purchased = result

        return EnVivoPageMedia(
            id=media_id,
            title=title,
            purchased=purchased,
        )

    def _get_envivo_live_metadata(
        self,
        stream_id: str,
        client_id: str,
    ) -> EnVivoMedia | None:
        log.debug("Requesting live metadata")

        payload = [
            {
                "operationName": "getPlayerLive",
                "variables": {
                    "id": stream_id,
                },
                "query": self._ENVIVO_LIVE_QUERY,
            },
        ]

        datatype, result = self._graphql(
            client_id,
            payload,
            {
                "data": {
                    "getLive": validate.any(
                        None,
                        validate.all(
                            {
                                "_id": str,
                                "name": str,
                                "accessToken": validate.any(str, None),
                                "schedules": [
                                    validate.all(
                                        {
                                            "name": str,
                                            "current": bool,
                                        },
                                        validate.union_get(
                                            "name",
                                            "current",
                                        ),
                                    ),
                                ],
                            },
                            validate.union_get(
                                "_id",
                                "name",
                                "accessToken",
                                "schedules",
                            ),
                        ),
                    ),
                },
            },
            validate.get(("data", "getLive")),
        )

        if datatype == "error":
            log.error(result)
            return None

        if result is None:
            log.error("This video is not available. Content may be geo-blocked")
            return None

        media_id, channel, access_token, schedules = result

        if not access_token:
            raise PluginError("This video is not available for playback")

        program = next(
            (name for name, current in schedules if current),
            None,
        )

        return EnVivoMedia(
            id=media_id,
            title=channel,
            channel=channel,
            program=program,
            access_token=access_token,
        )

    def _get_envivo_vod_metadata(
        self,
        stream_id: str,
        client_id: str,
    ) -> EnVivoMedia | None:
        page = self._get_envivo_page_metadata(stream_id, client_id)
        if page is None:
            return None

        if page.purchased < 1:
            log.warning(
                f"Unexpected purchased value ({page.purchased}); attempting playback anyway",
            )

        log.debug("Requesting VOD metadata")

        payload = [
            {
                "operationName": "getPlayerMedia",
                "variables": {
                    "withDescription": False,
                    "id": page.id,
                },
                "query": self._ENVIVO_MEDIA_QUERY,
            },
        ]

        datatype, result = self._graphql(
            client_id,
            payload,
            {
                "data": {
                    "getMedia": validate.any(
                        None,
                        validate.all(
                            {
                                "_id": str,
                                "title": str,
                                "accessToken": validate.any(str, None),
                            },
                            validate.union_get(
                                "_id",
                                "title",
                                "accessToken",
                            ),
                        ),
                    ),
                },
            },
            validate.get(("data", "getMedia")),
        )

        if datatype == "error":
            log.error(result)
            return None

        if result is None:
            log.error("This video is not available. Content may be geo-blocked")
            return None

        media_id, title, access_token = result

        if access_token is None:
            raise PluginError("This video is not available for playback")

        return EnVivoMedia(
            id=media_id,
            title=title,
            access_token=access_token,
        )

    def _get_envivo_config(self) -> EnVivoConfig:
        app_url = self.session.http.get(
            f"https://next.platform.mediastre.am/ott?domain={self._ENVIVO_DOMAIN}",
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "status": "OK",
                    "data": validate.url(
                        hostname="next-apps.mdstrm.com",
                    ),
                },
                validate.get("data"),
            ),
        )

        client_id, player_id, app_version = self.session.http.get(
            url_concat(app_url, "release.json"),
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "config": {
                        "api": {
                            "headers": {
                                "x-client-id": str,
                            },
                        },
                        "app": {
                            "version": str,
                        },
                        "players": {
                            "playerVod": str,
                        },
                    },
                },
                validate.union_get(
                    ("config", "api", "headers", "x-client-id"),
                    ("config", "players", "playerVod"),
                    ("config", "app", "version"),
                ),
            ),
        )

        return EnVivoConfig(
            client_id=client_id,
            player_id=player_id,
            app_version=app_version,
        )

    def _get_envivo_playback(
        self,
        endpoint: str,
        media_id: str,
        player_id: str,
        app_version: str,
        access_token: str,
    ) -> EnVivoPlayback:
        hls_url, uid, sid, pid = self.session.http.get(
            f"https://mdstrm.com/{endpoint}/{media_id}.json",
            params={
                "validate": "true",
                "metadata": "true",
                "access_token": access_token,
                "player": player_id,
                "language": "es",
                "an": self._ENVIVO_APP_NAME,
                "at": "web-app",
                "av": app_version,
                "ref": self._ENVIVO_DOMAIN,
                "res": self._ENVIVO_DEFAULT_RESOLUTION,
            },
            headers={
                "Referer": self._ENVIVO_REFERER,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "src": {
                        "hls": validate.url(),
                    },
                    "MDSTRMUID": str,
                    "MDSTRMSID": str,
                    "MDSTRMPID": str,
                },
                validate.union_get(
                    ("src", "hls"),
                    "MDSTRMUID",
                    "MDSTRMSID",
                    "MDSTRMPID",
                ),
            ),
        )

        return EnVivoPlayback(
            hls_url=hls_url,
            uid=uid,
            sid=sid,
            pid=pid,
        )

    def _graphql(
        self,
        client_id: str,
        payload: list[dict],
        *schemas,
    ):
        return self.session.http.post(
            "https://next.mediastream.co/graphql",
            headers={
                "Referer": self._ENVIVO_REFERER,
                "Origin": self._ENVIVO_ORIGIN,
                "x-client-id": client_id,
                "x-ott-language": "es",
            },
            json=payload,
            schema=validate.Schema(
                validate.parse_json(),
                [
                    validate.any(
                        validate.all(
                            {
                                "errors": [
                                    {
                                        "message": str,
                                    },
                                ],
                            },
                            validate.get("errors"),
                            validate.length(1),
                            validate.transform(
                                lambda errors: (
                                    "error",
                                    ", ".join(error["message"] for error in errors),
                                ),
                            ),
                        ),
                        validate.all(
                            dict,
                            *schemas,
                            validate.transform(lambda data: ("success", data)),
                        ),
                    ),
                ],
                validate.get(0),
            ),
        )


__plugin__ = TVAzteca



"""
$description Chinese live-streaming platform for live video game broadcasts, operated by Douyu Technology.
$url www.douyu.com
$type live
$metadata id
$metadata author
$metadata category
$metadata title
"""

from __future__ import annotations

import hashlib
import logging
import re
import sys
import time
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.session.http import TLSSecLevel1Adapter
from streamlink.stream.http import HTTPStream
from streamlink.utils.parse import parse_qsd


log = logging.getLogger(__name__)


class TLSPartialVerifyAdapter(TLSSecLevel1Adapter):
    """
    The base class SSLContextAdapter resets ssl_context.check_hostname to True when sending a request,
    so overriding get_ssl_context does not take effect.

    Instead, assert_hostname is used to influence urllib3's internal decision logic,
    ultimately achieving the effect of check_hostname=False to skip certificate hostname verification,
    while still retaining certificate chain-related validation."
    """

    def init_poolmanager(self, *args, **kwargs):
        kwargs["assert_hostname"] = False
        return super().init_poolmanager(*args, **kwargs)


@pluginmatcher(
    re.compile(
        r"https?://(?:www\.)?douyu\.com/(?:topic/)?(?P<rid>\d+)",
    ),
)
class Douyu(Plugin):
    _URL_ENCRYPTION = "https://www.douyu.com/wgapi/livenc/liveweb/websec/getEncryption"
    _URL_PLAY = "https://www.douyu.com/lapi/live/getH5PlayV1/{rid}"
    _URL_BETARD = "https://www.douyu.com/betard/{rid}"

    DID = "10000000000000000000000000001501"
    HEVC = "0"
    FA = "0"
    IVE = "0"

    QUALITY_WEIGHTS: dict[str, int] = {}

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.QUALITY_WEIGHTS.get(stream)
        if weight:
            return weight, "douyu"
        return super().stream_weight(stream)

    def _get_room_metadata(self, rid: str) -> bool:
        try:
            data = self.session.http.get(
                self._URL_BETARD.format(rid=rid),
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "room": {
                            "room_id": int,
                            "room_name": str,
                            "nickname": str,
                            validate.optional("cate_name"): validate.any(None, str),
                            "show_status": int,
                        },
                    },
                    validate.get("room"),
                ),
            )
            self.id = str(data["room_id"])
            self.title = data["room_name"]
            self.author = data["nickname"]
            self.category = data.get("cate_name")
            return data["show_status"] == 1
        except Exception:
            log.debug("Failed to get room metadata")
            return True

    @staticmethod
    def _compute_auth(rid: str, ts: int, key: str, rand_str: str, enc_time: int, is_special: int) -> str:
        suffix = "" if is_special == 1 else f"{rid}{ts}"
        f = rand_str
        for _ in range(enc_time):
            f = hashlib.md5((f + key).encode("utf-8")).hexdigest()
        return hashlib.md5((f + key + suffix).encode("utf-8")).hexdigest()

    def _get_encryption(self, did: str) -> dict | None:
        enc_response = self.session.http.get(
            self._URL_ENCRYPTION,
            params={"did": did},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "error": int,
                    "data": validate.any(
                        None,
                        {
                            "key": str,
                            "rand_str": str,
                            "enc_time": int,
                            "enc_data": str,
                            "is_special": int,
                        },
                    ),
                },
            ),
        )
        if enc_response["error"] != 0 or not enc_response["data"]:
            return None
        return enc_response["data"]

    def _request_stream(self, rid: str, rate: int, did: str) -> dict | None:
        enc_data = self._get_encryption(did)
        if not enc_data:
            log.error("Failed to get encryption parameters")
            return None

        ts = int(time.time())
        auth = self._compute_auth(
            rid=rid,
            ts=ts,
            key=enc_data["key"],
            rand_str=enc_data["rand_str"],
            enc_time=enc_data["enc_time"],
            is_special=enc_data["is_special"],
        )

        play_data = self.session.http.post(
            self._URL_PLAY.format(rid=rid),
            data={
                "enc_data": enc_data["enc_data"],
                "tt": str(ts),
                "did": did,
                "auth": auth,
                "cdn": "",
                "rate": str(rate),
                "hevc": self.HEVC,
                "fa": self.FA,
                "ive": self.IVE,
            },
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "error": int,
                    "data": validate.any(
                        None,
                        "",
                        {
                            "rtmp_url": validate.url(scheme="https"),
                            "rtmp_live": str,
                            validate.optional("multirates"): [{"name": str, "rate": int, "bit": int}],
                        },
                    ),
                },
            ),
        )

        if play_data["error"] != 0 or not play_data["data"]:
            return None

        return play_data["data"]

    def _get_streams(self):
        rid = self.match.group("rid")

        params = parse_qsd(urlparse(self.url).query)
        did = params.get("dyshid") or self.DID

        if not self._get_room_metadata(rid):
            log.info("Streamer is offline")
            return

        first_data = self._request_stream(rid, 0, did)
        if not first_data:
            log.error("Failed to get stream URL")
            return

        # Mount TLSPartialVerifyAdapter for CDN domains that may not be compatible
        # with OpenSSL 3.0's default security level (SECLEVEL=2)
        # or domains with underscore
        self.session.http.mount("https://", TLSPartialVerifyAdapter())

        multirates = first_data.get("multirates", [])

        if not multirates:
            url = f"{first_data['rtmp_url']}/{first_data['rtmp_live']}"
            self.QUALITY_WEIGHTS["source"] = sys.maxsize
            yield (
                "source",
                HTTPStream(
                    self.session,
                    url,
                    headers={"Referer": "https://www.douyu.com/"},
                ),
            )
            return

        for rate_info in multirates:
            rate = rate_info["rate"]
            bit = rate_info["bit"]
            name = f"{bit}k" if bit else "source"
            weight = bit or sys.maxsize
            self.QUALITY_WEIGHTS[name] = weight

            if rate == 0:
                data = first_data
            else:
                data = self._request_stream(rid, rate, did)

            if data:
                url = f"{data['rtmp_url']}/{data['rtmp_live']}"
                yield (
                    name,
                    HTTPStream(
                        self.session,
                        url,
                        headers={"Referer": "https://www.douyu.com/"},
                    ),
                )

        log.debug("QUALITY_WEIGHTS: %r", self.QUALITY_WEIGHTS)


__plugin__ = Douyu



import socket
import ssl
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, TypedDict, overload

# noinspection PyProtectedMember
import requests._types as _requeststypes  # ruff: ignore[import-private-name]
from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter
from requests.cookies import CookieJar, RequestsCookieJar
from typing_extensions import Unpack

from streamlink.plugin.api.validate import Schema

class _StreamlinkKwargs(TypedDict, total=False):
    acceptable_status: Sequence[int] | None
    encoding: str | None
    exception: type[Exception]
    raise_for_status: bool
    retries: float
    retry_backoff: float
    retry_max_backoff: float
    session: HTTPSession

class _SchemaKwargs(TypedDict):
    schema: Schema

class _RequestKwargs(_requeststypes.RequestKwargs, _StreamlinkKwargs):
    pass

class _RequestKwargsWithSchema(_requeststypes.RequestKwargs, _StreamlinkKwargs, _SchemaKwargs):
    pass

class _GetKwargs(_requeststypes.GetKwargs, _StreamlinkKwargs):
    pass

class _GetKwargsWithSchema(_requeststypes.GetKwargs, _StreamlinkKwargs, _SchemaKwargs):
    pass

class _PostKwargs(_requeststypes.PostKwargs, _StreamlinkKwargs):
    pass

class _PostKwargsWithSchema(_requeststypes.PostKwargs, _StreamlinkKwargs, _SchemaKwargs):
    pass

class _DataKwargs(_requeststypes.DataKwargs, _StreamlinkKwargs):
    pass

class _DataKwargsWithSchema(_requeststypes.DataKwargs, _StreamlinkKwargs, _SchemaKwargs):
    pass

# ----

def urllib3_set_socket_options(sock: socket.socket, options: list[tuple[int, int, int | bytes]] | None) -> None: ...

class SSLContextAdapter(HTTPAdapter):
    def get_ssl_context(self) -> ssl.SSLContext: ...

class TLSNoDHAdapter(SSLContextAdapter): ...
class TLSSecLevel1Adapter(SSLContextAdapter): ...

class HTTPSession(Session):
    params: dict
    timeout: float

    @classmethod
    def determine_json_encoding(cls, sample: bytes) -> str: ...
    @classmethod
    def json(
        cls,
        res: Response,
        name: str | None = ...,
        exception: type[Exception] | None = ...,
        schema: Schema | None = ...,
        *args,
        **kwargs,
    ) -> Any: ...
    @classmethod
    def xml(
        cls,
        res: Response,
        ignore_ns: bool | None = ...,
        invalid_char_entities: bool | None = ...,
        name: str | None = ...,
        exception: type[Exception] | None = ...,
        schema: Schema | None = ...,
        *args,
        **kwargs,
    ) -> Any: ...
    def set_interface(self, interface: str | None) -> None: ...
    def set_address_family(self, family: socket.AddressFamily | None = None) -> None: ...
    def disable_dh(self, disable: bool = True) -> None: ...
    def set_cookies_from_file(self, path: Path | str) -> None: ...
    def resolve_url(self, url: str) -> str: ...
    @staticmethod
    def valid_request_args(**req_keywords) -> dict[str, Any]: ...
    def prepare_new_request(self, **req_keywords) -> PreparedRequest: ...
    @overload
    def request(
        self,
        method: str,
        url: _requeststypes.UriType,
        params: _requeststypes.ParamsType = None,
        data: _requeststypes.DataType = None,
        headers: Mapping[str, str | bytes] | None = None,
        cookies: RequestsCookieJar | CookieJar | dict[str, str] | None = None,
        files: _requeststypes.FilesType = None,
        auth: _requeststypes.AuthType = None,
        timeout: _requeststypes.TimeoutType = None,
        allow_redirects: bool = True,
        proxies: dict[str, str] | None = None,
        hooks: _requeststypes.HooksInputType | None = None,
        stream: bool | None = None,
        verify: _requeststypes.VerifyType | None = None,
        cert: _requeststypes.CertType = None,
        json: _requeststypes.JsonType = None,
        # streamlink options
        acceptable_status: Sequence[int] | None = None,
        encoding: str | None = None,
        exception: type[Exception] | None = None,
        raise_for_status: bool = True,
        retries: int = 0,
        retry_backoff: float = 0.3,
        retry_max_backoff: float = 10.0,
        schema: Schema | None = None,
        session: HTTPSession | None = None,
    ) -> Any: ...
    @overload
    def request(
        self,
        url: _requeststypes.UriType,
        params: _requeststypes.ParamsType = None,
        data: _requeststypes.DataType = None,
        headers: Mapping[str, str | bytes] | None = None,
        cookies: RequestsCookieJar | CookieJar | dict[str, str] | None = None,
        files: _requeststypes.FilesType = None,
        auth: _requeststypes.AuthType = None,
        timeout: _requeststypes.TimeoutType = None,
        allow_redirects: bool = True,
        proxies: dict[str, str] | None = None,
        hooks: _requeststypes.HooksInputType | None = None,
        stream: bool | None = None,
        verify: _requeststypes.VerifyType | None = None,
        cert: _requeststypes.CertType = None,
        json: _requeststypes.JsonType = None,
        # streamlink options
        acceptable_status: Sequence[int] | None = None,
        encoding: str | None = None,
        exception: type[Exception] | None = None,
        raise_for_status: bool = True,
        retries: int = 0,
        retry_backoff: float = 0.3,
        retry_max_backoff: float = 10.0,
        session: HTTPSession | None = None,
    ) -> Response: ...
    @overload
    def get(
        self,
        url: _requeststypes.UriType,
        params: _requeststypes.ParamsType | None = None,
        **kwargs: Unpack[_GetKwargsWithSchema],
    ) -> Any: ...
    @overload
    def get(
        self,
        url: _requeststypes.UriType,
        params: _requeststypes.ParamsType | None = None,
        **kwargs: Unpack[_GetKwargs],
    ) -> Response: ...
    @overload
    def options(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargsWithSchema],
    ) -> Any: ...
    @overload
    def options(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargs],
    ) -> Response: ...
    @overload
    def head(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargsWithSchema],
    ) -> Any: ...
    @overload
    def head(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargs],
    ) -> Response: ...
    @overload
    def post(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        json: _requeststypes.JsonType = None,
        **kwargs: Unpack[_PostKwargsWithSchema],
    ) -> Any: ...
    @overload
    def post(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        json: _requeststypes.JsonType = None,
        **kwargs: Unpack[_PostKwargs],
    ) -> Response: ...
    @overload
    def put(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        **kwargs: Unpack[_DataKwargsWithSchema],
    ) -> Any: ...
    @overload
    def put(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        **kwargs: Unpack[_DataKwargs],
    ) -> Response: ...
    @overload
    def patch(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        **kwargs: Unpack[_DataKwargsWithSchema],
    ) -> Any: ...
    @overload
    def patch(
        self,
        url: _requeststypes.UriType,
        data: _requeststypes.DataType = None,
        **kwargs: Unpack[_DataKwargs],
    ) -> Response: ...
    @overload
    def delete(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargsWithSchema],
    ) -> Any: ...
    @overload
    def delete(
        self,
        url: _requeststypes.UriType,
        **kwargs: Unpack[_RequestKwargs],
    ) -> Response: ...



"""
$description Chinese live-streaming platform owned by ByteDance.
$url live.douyin.com
$type live
$metadata id
$metadata author
$metadata title
"""

from __future__ import annotations

import re
import sys
import uuid

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(
        r"https?://(?:live\.)?douyin\.com/(?P<room_id>[^/?]+)",
    ),
)
class Douyin(Plugin):
    _STATUS_LIVE = 2

    # There's no bitrate information in the JSON data available anymore.
    # These weights are based on empirically measured video bitrates of a specific stream during plugin fixing.
    # The values themselves don't matter, only the order.
    QUALITY_WEIGHTS: dict[str, int] = {
        "full_hd1": sys.maxsize,
        "hd1": 8000,
        "sd2": 2000,
        "sd1": 1000,
    }

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.QUALITY_WEIGHTS.get(stream)
        if weight:
            return weight, stream

        return super().stream_weight(stream)

    SCHEMA_ROOM_STORE = validate.all(
        {
            "roomInfo": {
                validate.optional("room"): validate.all(
                    {
                        "status": int,
                        "id_str": str,
                        "title": str,
                        validate.optional("owner"): validate.all(
                            {"nickname": str},
                            validate.get("nickname"),
                        ),
                        validate.optional("stream_url"): {
                            "flv_pull_url": {
                                str: validate.url(),
                            },
                        },
                    },
                    validate.union_get(
                        "status",
                        "id_str",
                        "title",
                        "owner",
                        "stream_url",
                    ),
                ),
            },
        },
        validate.get(("roomInfo", "room")),
    )

    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            cookies={
                "__ac_nonce": uuid.uuid4().hex[:21],
            },
            schema=validate.Schema(
                validate.regex(
                    pattern=re.compile(r"self\.__pace_f\.push\(\[\d+,(\"\w+:.+?\")]\)</script>"),
                    method="findall",
                ),
                validate.filter(lambda item: "state" in item and "streamStore" in item),
                validate.get(-1),
                validate.none_or_all(
                    validate.parse_json(),
                    validate.transform(lambda s: re.sub(r"^\w+:", "", s)),
                    validate.parse_json(),
                    list,
                    validate.filter(lambda item: isinstance(item, dict) and "state" in item),
                    validate.length(1),
                    validate.get((0, "state")),
                    {
                        "roomStore": self.SCHEMA_ROOM_STORE,
                    },
                    validate.get("roomStore"),
                ),
            ),
        )
        if not data:
            return

        status, self.id, self.title, self.author, streams = data
        if status != self._STATUS_LIVE:
            log.info("The channel is currently offline")
            return

        for name, flv_url in streams["flv_pull_url"].items():
            yield name.lower(), HTTPStream(self.session, update_scheme("https://", flv_url, force=True))


__plugin__ = Douyin



from __future__ import annotations

import math
from typing import Any, Literal, overload


@overload
def to_float(
    value: Any,
    *,
    allow_nan: bool = False,
    allow_inf: bool = False,
) -> float | None: ...  # pragma: no cover


@overload
def to_float(
    value: Any,
    *,
    allow_nan: bool = False,
    allow_inf: bool = False,
    raise_on_error: Literal[True] = True,
) -> float: ...  # pragma: no cover


@overload
def to_float(
    value: Any,
    *,
    allow_nan: bool = False,
    allow_inf: bool = False,
    raise_on_error: bool = False,
) -> float | None: ...  # pragma: no cover


def to_float(value: Any, *, allow_nan: bool = False, allow_inf: bool = False, raise_on_error: bool = False) -> float | None:
    try:
        num = float(value)
    except (ValueError, TypeError, OverflowError):
        if raise_on_error:
            raise
        return None

    if not allow_nan and math.isnan(num):
        if raise_on_error:
            raise ValueError("NaN values are not allowed")
        return None

    if not allow_inf and math.isinf(num):
        if raise_on_error:
            raise ValueError("Infinity values are not allowed")
        return None

    return num



"""
$description Live TV channels and video on-demand service from NOS, a Dutch public, state-owned broadcaster.
$url nos.nl
$type live, vod
$metadata id
$metadata title
$region Netherlands
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:\w+\.)?nos\.nl/(?:live$|livestream/|l/|video/|collectie/)"),
)
class NOS(Plugin):
    def _get_streams(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[@type='application/ld+json'][1]/text()"),
                validate.none_or_all(
                    validate.parse_json(),
                    {
                        "@graph": validate.all(
                            [dict],
                            validate.filter(lambda item: item.get("@type") == "VideoObject"),
                            validate.filter(lambda item: item.get("encodingFormat") == "application/vnd.apple.mpegurl"),
                        ),
                    },
                    validate.get("@graph"),
                    validate.get(0),
                    validate.none_or_all(
                        {
                            "contentUrl": validate.url(),
                            "identifier": validate.any(int, str),
                            "name": str,
                        },
                        validate.union_get(
                            "contentUrl",
                            "identifier",
                            "name",
                        ),
                    ),
                ),
            ),
        )
        if not data:
            return

        hls_url, self.id, self.title = data

        headers = {
            "Origin": "https://nos.nl",
            "Referer": self.url,
        }
        res = self.session.http.get(hls_url, headers=headers, raise_for_status=False)
        if res.status_code >= 400:
            log.error("Content is inaccessible or may have expired")
            return

        return HLSStream.parse_variant_playlist(self.session, hls_url, headers=headers)


__plugin__ = NOS
